@@ -3,7 +3,7 @@
 #include <dirent.h>
 #include <errno.h>
 //#include <sys/prctl.h>
-#include <sys/types.h>
+//#include <sys/types.h>
 #include <unistd.h>
 #include <stdio.h>
 
@@ -15,12 +15,12 @@
 #include "escape.h"
 #include "exec-util.h"
 #include "fd-util.h"
-#include "fileio.h"
+//#include "fileio.h"
 #include "hashmap.h"
 #include "macro.h"
-//#include "missing_syscall.h"
+#include "missing_syscall.h"
+#include "path-util.h"
 #include "process-util.h"
-#include "rlimit-util.h"
 #include "serialize.h"
 //#include "set.h"
 //#include "signal-util.h"
@@ -29,13 +29,18 @@
 #include "string-util.h"
 #include "strv.h"
 //#include "terminal-util.h"
-//#include "tmpfile-util.h"
-//#include "util.h"
+#include "tmpfile-util.h"
+
+#define EXIT_SKIP_REMAINING 77
 
 /* Put this test here for a lack of better place */
 assert_cc(EAGAIN == EWOULDBLOCK);
 
+#if 0 /// elogind is never INIT and therefore never sets SYSTEMD_EXEC_PID
 static int do_spawn(const char *path, char *argv[], int stdout_fd, pid_t *pid, bool set_systemd_exec_pid) {
+#else // 0
+static int do_spawn(const char *path, char *argv[], int stdout_fd, pid_t *pid) {
+#endif // 0
         pid_t _pid;
         int r;
 
@@ -44,7 +49,7 @@ static int do_spawn(const char *path, char *argv[], int stdout_fd, pid_t *pid, b
                 return 0;
         }
 
-        r = safe_fork("(direxec)", FORK_DEATHSIG|FORK_LOG, &_pid);
+        r = safe_fork("(direxec)", FORK_DEATHSIG_SIGTERM|FORK_LOG|FORK_RLIMIT_NOFILE_SAFE, &_pid);
         if (r < 0)
                 return r;
         if (r == 0) {
@@ -56,13 +61,13 @@ static int do_spawn(const char *path, char *argv[], int stdout_fd, pid_t *pid, b
                                 _exit(EXIT_FAILURE);
                 }
 
-                (void) rlimit_nofile_safe();
-
+#if 0 /// elogind is never INIT and therefore never sets SYSTEMD_EXEC_PID
                 if (set_systemd_exec_pid) {
-                        r = setenv_elogind_exec_pid(false);
+                        r = setenv_systemd_exec_pid(false);
                         if (r < 0)
                                 log_warning_errno(r, "Failed to set $SYSTEMD_EXEC_PID, ignoring: %m");
                 }
+#endif // 0
 
                 if (!argv) {
                         _argv[0] = (char*) path;
@@ -81,7 +86,8 @@ static int do_spawn(const char *path, char *argv[], int stdout_fd, pid_t *pid, b
 }
 
 static int do_execute(
-                char **directories,
+                char* const* paths,
+                const char *root,
                 usec_t timeout,
                 gather_stdout_callback_t const callbacks[_STDOUT_CONSUME_MAX],
                 void* const callback_args[_STDOUT_CONSUME_MAX],
@@ -91,9 +97,8 @@ static int do_execute(
                 ExecDirFlags flags) {
 
         _cleanup_hashmap_free_free_ Hashmap *pids = NULL;
-        _cleanup_strv_free_ char **paths = NULL;
-        int r;
         bool parallel_execution;
+        int r;
 
         /* We fork this all off from a child process so that we can somewhat cleanly make
          * use of SIGALRM to set a time limit.
@@ -103,10 +108,6 @@ static int do_execute(
          */
         parallel_execution = FLAGS_SET(flags, EXEC_DIR_PARALLEL) && !callbacks;
 
-        r = conf_files_list_strv(&paths, NULL, NULL, CONF_FILES_EXECUTABLE|CONF_FILES_REGULAR|CONF_FILES_FILTER_MASKED, (const char* const*) directories);
-        if (r < 0)
-                return log_error_errno(r, "Failed to enumerate executables: %m");
-
         if (parallel_execution) {
                 pids = hashmap_new(NULL);
                 if (!pids)
@@ -125,19 +126,25 @@ static int do_execute(
 
         STRV_FOREACH(path, paths) {
                 _cleanup_free_ char *t = NULL;
-                _cleanup_close_ int fd = -1;
+                _cleanup_close_ int fd = -EBADF;
 #if 0 /// No "maybe uninitialized" warning in elogind
                 pid_t pid;
 #else // 0
                 pid_t pid = 0;
 #endif // 0
 
-                t = strdup(*path);
+                t = path_join(root, *path);
                 if (!t)
                         return log_oom();
 
                 if (callbacks) {
-                        fd = open_serialization_fd(basename(*path));
+                        _cleanup_free_ char *bn = NULL;
+
+                        r = path_extract_filename(*path, &bn);
+                        if (r < 0)
+                                return log_error_errno(r, "Failed to extract filename from path '%s': %m", *path);
+
+                        fd = open_serialization_fd(bn);
                         if (fd < 0)
                                 return log_error_errno(fd, "Failed to open serialization file: %m");
                 }
@@ -150,7 +157,11 @@ static int do_execute(
                         log_debug("About to execute %s%s%s", t, argv ? " " : "", argv ? strnull(args) : "");
                 }
 
+#if 0 /// elogind never sets this PID as it is never INIT
                 r = do_spawn(t, argv, fd, &pid, FLAGS_SET(flags, EXEC_DIR_SET_SYSTEMD_EXEC_PID));
+#else // 0
+                r = do_spawn(t, argv, fd, &pid);
+#endif // 0
                 if (r <= 0)
                         continue;
 
@@ -160,22 +171,34 @@ static int do_execute(
                                 return log_oom();
                         t = NULL;
                 } else {
-                        r = wait_for_terminate_and_check(t, pid, WAIT_LOG);
-                        if (FLAGS_SET(flags, EXEC_DIR_IGNORE_ERRORS)) {
-                                if (r < 0)
-                                        continue;
-                        } else if (r > 0)
+                        bool skip_remaining = false;
+
+                        r = wait_for_terminate_and_check(t, pid, WAIT_LOG_ABNORMAL);
+                        if (r < 0)
                                 return r;
+                        if (r > 0) {
+                                if (FLAGS_SET(flags, EXEC_DIR_SKIP_REMAINING) && r == EXIT_SKIP_REMAINING) {
+                                        log_info("%s succeeded with exit status %i, not executing remaining executables.", *path, r);
+                                        skip_remaining = true;
+                                } else if (FLAGS_SET(flags, EXEC_DIR_IGNORE_ERRORS))
+                                        log_warning("%s failed with exit status %i, ignoring.", *path, r);
+                                else {
+                                        log_error("%s failed with exit status %i.", *path, r);
+                                        return r;
+                                }
+                        }
 
                         if (callbacks) {
                                 if (lseek(fd, 0, SEEK_SET) < 0)
                                         return log_error_errno(errno, "Failed to seek on serialization fd: %m");
 
-                                r = callbacks[STDOUT_GENERATE](fd, callback_args[STDOUT_GENERATE]);
-                                fd = -1;
+                                r = callbacks[STDOUT_GENERATE](TAKE_FD(fd), callback_args[STDOUT_GENERATE]);
                                 if (r < 0)
                                         return log_error_errno(r, "Failed to process output from %s: %m", *path);
                         }
+
+                        if (skip_remaining)
+                                break;
                 }
         }
 
@@ -196,6 +219,8 @@ static int do_execute(
                 assert(t);
 
                 r = wait_for_terminate_and_check(t, pid, WAIT_LOG);
+                if (r < 0)
+                        return r;
                 if (!FLAGS_SET(flags, EXEC_DIR_IGNORE_ERRORS) && r > 0)
                         return r;
         }
@@ -203,8 +228,10 @@ static int do_execute(
         return 0;
 }
 
-int execute_directories(
-                const char* const* directories,
+int execute_strv(
+                const char *name,
+                char* const* paths,
+                const char *root,
                 usec_t timeout,
                 gather_stdout_callback_t const callbacks[_STDOUT_CONSUME_MAX],
                 void* const callback_args[_STDOUT_CONSUME_MAX],
@@ -212,18 +239,17 @@ int execute_directories(
                 char *envp[],
                 ExecDirFlags flags) {
 
-        char **dirs = (char**) directories;
-        _cleanup_close_ int fd = -1;
-        char *name;
-        int r;
+        _cleanup_close_ int fd = -EBADF;
         pid_t executor_pid;
+        int r;
 
-        assert(!strv_isempty(dirs));
+        assert(!FLAGS_SET(flags, EXEC_DIR_PARALLEL | EXEC_DIR_SKIP_REMAINING));
 
-        name = basename(dirs[0]);
-        assert(!isempty(name));
+        if (strv_isempty(paths))
+                return 0;
 
         if (callbacks) {
+                assert(name);
                 assert(callback_args);
                 assert(callbacks[STDOUT_GENERATE]);
                 assert(callbacks[STDOUT_COLLECT]);
@@ -238,15 +264,15 @@ int execute_directories(
          * them to finish. Optionally a timeout is applied. If a file with the same name
          * exists in more than one directory, the earliest one wins. */
 
-        r = safe_fork("(sd-executor)", FORK_RESET_SIGNALS|FORK_DEATHSIG|FORK_LOG, &executor_pid);
+        r = safe_fork("(sd-exec-strv)", FORK_RESET_SIGNALS|FORK_DEATHSIG_SIGTERM|FORK_LOG, &executor_pid);
         if (r < 0)
                 return r;
         if (r == 0) {
-                r = do_execute(dirs, timeout, callbacks, callback_args, fd, argv, envp, flags);
+                r = do_execute(paths, root, timeout, callbacks, callback_args, fd, argv, envp, flags);
                 _exit(r < 0 ? EXIT_FAILURE : r);
         }
 
-        r = wait_for_terminate_and_check("(sd-executor)", executor_pid, 0);
+        r = wait_for_terminate_and_check("(sd-exec-strv)", executor_pid, 0);
         if (r < 0)
                 return r;
         if (!FLAGS_SET(flags, EXEC_DIR_IGNORE_ERRORS) && r > 0)
@@ -258,13 +284,45 @@ int execute_directories(
         if (lseek(fd, 0, SEEK_SET) < 0)
                 return log_error_errno(errno, "Failed to rewind serialization fd: %m");
 
-        r = callbacks[STDOUT_CONSUME](fd, callback_args[STDOUT_CONSUME]);
-        fd = -1;
+        r = callbacks[STDOUT_CONSUME](TAKE_FD(fd), callback_args[STDOUT_CONSUME]);
         if (r < 0)
                 return log_error_errno(r, "Failed to parse returned data: %m");
         return 0;
 }
 
+int execute_directories(
+                const char* const* directories,
+                usec_t timeout,
+                gather_stdout_callback_t const callbacks[_STDOUT_CONSUME_MAX],
+                void* const callback_args[_STDOUT_CONSUME_MAX],
+                char *argv[],
+                char *envp[],
+                ExecDirFlags flags) {
+
+        _cleanup_strv_free_ char **paths = NULL;
+        _cleanup_free_ char *name = NULL;
+        int r;
+
+        assert(!strv_isempty((char**) directories));
+
+        r = conf_files_list_strv(&paths, NULL, NULL, CONF_FILES_EXECUTABLE|CONF_FILES_REGULAR|CONF_FILES_FILTER_MASKED, directories);
+        if (r < 0)
+                return log_error_errno(r, "Failed to enumerate executables: %m");
+
+        if (strv_isempty(paths)) {
+                log_debug("No executables found.");
+                return 0;
+        }
+
+        if (callbacks) {
+                r = path_extract_filename(directories[0], &name);
+                if (r < 0)
+                        return log_error_errno(r, "Failed to extract file name from '%s': %m", directories[0]);
+        }
+
+        return execute_strv(name, paths, NULL, timeout, callbacks, callback_args, argv, envp, flags);
+}
+
 #if 0 /// UNNEEDED by elogind
 static int gather_environment_generate(int fd, void *arg) {
         char ***env = ASSERT_PTR(arg);
@@ -395,14 +453,14 @@ int exec_command_flags_to_strv(ExecCommandFlags flags, char ***ex_opts) {
         _cleanup_strv_free_ char **ret_opts = NULL;
         ExecCommandFlags it = flags;
         const char *str;
-        int i, r;
+        int r;
 
         assert(ex_opts);
 
         if (flags < 0)
                 return flags;
 
-        for (i = 0; it != 0; it &= ~(1 << i), i++) {
+        for (unsigned i = 0; it != 0; it &= ~(1 << i), i++)
                 if (FLAGS_SET(flags, (1 << i))) {
                         str = exec_command_flags_to_string(1 << i);
                         if (!str)
@@ -412,7 +470,6 @@ int exec_command_flags_to_strv(ExecCommandFlags flags, char ***ex_opts) {
                         if (r < 0)
                                 return r;
                 }
-        }
 
         *ex_opts = TAKE_PTR(ret_opts);
 
@@ -434,9 +491,7 @@ static const char* const exec_command_strings[] = {
 };
 
 const char* exec_command_flags_to_string(ExecCommandFlags i) {
-        size_t idx;
-
-        for (idx = 0; idx < ELEMENTSOF(exec_command_strings); idx++)
+        for (size_t idx = 0; idx < ELEMENTSOF(exec_command_strings); idx++)
                 if (i == (1 << idx))
                         return exec_command_strings[idx];
 
@@ -499,9 +554,10 @@ int fork_agent(const char *name, const int except[], size_t n_except, pid_t *ret
         /* Spawns a temporary TTY agent, making sure it goes away when we go away */
 
         r = safe_fork_full(name,
+                           NULL,
                            except,
                            n_except,
-                           FORK_RESET_SIGNALS|FORK_DEATHSIG|FORK_CLOSE_ALL_FDS|FORK_REOPEN_LOG,
+                           FORK_RESET_SIGNALS|FORK_DEATHSIG_SIGTERM|FORK_CLOSE_ALL_FDS|FORK_REOPEN_LOG|FORK_RLIMIT_NOFILE_SAFE,
                            ret_pid);
         if (r < 0)
                 return r;
@@ -545,8 +601,6 @@ int fork_agent(const char *name, const int except[], size_t n_except, pid_t *ret
                 }
         }
 
-        (void) rlimit_nofile_safe();
-
         /* Count arguments */
         va_start(ap, path);
         for (n = 0; va_arg(ap, char*); n++)
